package composer;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import metadata.CompositionMetadataStore;
import modification.content.InvalidFSTTraversalException;
import modification.traversalLanguageParser.ParseException;
import printer.PrintVisitorException;
import builder.ArtifactBuilderInterface;
import builder.java.JavaBuilder;

import composer.rules.CSharpMethodOverriding;
import composer.rules.CompositionError;
import composer.rules.CompositionRule;
import composer.rules.ConstructorConcatenation;
import composer.rules.ExpansionOverriding;
import composer.rules.FieldOverriding;
import composer.rules.ImplementsListMerging;
import composer.rules.JavaMethodOverriding;
import composer.rules.ModifierListSpecialization;
import composer.rules.Replacement;
import composer.rules.StringConcatenation;
import composer.rules.rtcomp.RuntimeFeatureSelection;
import composer.rules.rtcomp.RuntimeFunctionRefinement;
import composer.rules.rtcomp.RuntimeReplacement;
import composer.rules.rtcomp.RuntimeSubtreeIntegration;
import counter.Counter;

import de.ovgu.cide.fstgen.ast.AbstractFSTParser;
import de.ovgu.cide.fstgen.ast.FSTNode;
import de.ovgu.cide.fstgen.ast.FSTNonTerminal;
import de.ovgu.cide.fstgen.ast.FSTTerminal;
import de.ovgu.cide.fstgen.ast.FSTVisitor;

public class FSTGenComposer extends FSTGenProcessor {

	private CmdLineInterpreter cmd = new CmdLineInterpreter();

	private CompositionMetadataStore meta = CompositionMetadataStore.getInstance();
	private List<CompositionRule> compositionRules;
	private RuntimeSubtreeIntegration subtreeRewriter = null;
	
	public FSTGenComposer() {
		super();
		
	}
	
	private FSTNode rewriteSubtree(FSTNode n) {
		meta.discoverFuncIntroductions(n);	
		if (cmd.lifting) {
			return subtreeRewriter.rewrite(n.getDeepClone());
		} else {
			return n.getDeepClone();
		}
	}
	
	public void run(String[] args) {
		cmd.parseCmdLineArguments(args);
		
		//select the composition rules
		compositionRules = new ArrayList<CompositionRule>();
		
		if (cmd.lifting) {			
			compositionRules.add(new RuntimeReplacement());
			compositionRules.add(new RuntimeFunctionRefinement());			

			subtreeRewriter = new RuntimeSubtreeIntegration();
		} else {
			compositionRules.add(new Replacement());
			compositionRules.add(new JavaMethodOverriding());

		}
		
		try {
			try {
				fileLoader.loadFiles(cmd.equationFileName, cmd.equationBaseDirectoryName, cmd.isAheadEquationFile);
			} catch (cide.gparser.ParseException e1) {
				System.out.println("error");
				fireParseErrorOccured(e1);
				e1.printStackTrace();
			}
	
			String outputDir = cmd.equationBaseDirectoryName;
			if (cmd.outputDirectoryName != null)
				outputDir = cmd.outputDirectoryName;

			featureVisitor.setWorkingDir(outputDir);
			
			featureVisitor.setExpressionName(cmd.equationFileName);
			
			for (ArtifactBuilderInterface builder : getArtifactBuilders()) {
				
				System.out.println(builder.getClass().getCanonicalName());
				
				LinkedList<FSTNonTerminal> features = builder.getFeatures();

				if(cmd.isCount && builder instanceof JavaBuilder) {
					Counter counter = new Counter();
					for (FSTNonTerminal feature : features) {
						counter.collect(feature);
					}
					if(features.size() > 0)
						counter.writeFile(new File(cmd.equationFileName + ".rsf"));
				}
				System.out.println("SELECTED FEATURES:");
				
				for (FSTNonTerminal feature : features) {
					meta.addFeature(feature.getName());

				}
				
				FSTNode composition = compose(features);
//				modify(composition);

//				if(composition != null)
//					System.err.println(composition.toString());
				
				
				/* 
				 * hook for general purpose visitors
				 */
				if (null != composition)
        				for (FSTVisitor visitor: getFSTVisitors()) {
        				    composition.accept(visitor);
        				}
				
				try {
					featureVisitor.visit((FSTNonTerminal) composition);
				} catch (PrintVisitorException e) {
					e.printStackTrace();
				}
			}
			setFstnodes(AbstractFSTParser.fstnodes);
			System.out.println(meta);
			try {
				//System.out.println(outputDir + "features/roles.meta");
				
				String exp = new File(cmd.equationFileName).getName();
				exp = exp.substring(0, exp.length() - 4);
				
				meta.saveToFile(outputDir + "/" + exp + "/roles.meta");
				if (cmd.lifting) {
					File cnfFile = new File(cmd.equationBaseDirectoryName, "model.cnf");
					System.err.println("cnfFile:" + cnfFile.getAbsolutePath());
					new RuntimeFeatureSelection(meta, cnfFile).saveTo(outputDir + "features/featureselect");
				}
			} catch (IOException e) {			
				e.printStackTrace();
			}
		} catch (FileNotFoundException e1) {
			//e1.printStackTrace();
		}

	}

	private void modify(FSTNode composition) {
		if (composition != null) {
			try {
				fileLoader.getModifications().apply(composition);
			} catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (ParseException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (cide.gparser.ParseException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (InvalidFSTTraversalException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			System.out.println(composition);
		}
	}

	public static void main(String[] args) {
		FSTGenComposer composer = new FSTGenComposer();
		composer.run(args);
	}

	private FSTNode compose(List<FSTNonTerminal> tl) {
		FSTNode composed = null;
		for (FSTNode current : tl) {
			if (composed != null) {
				composed = compose(current, composed);
			} else
				composed = current;
		}
		return composed;
	}

	public FSTNode compose(FSTNode nodeA, FSTNode nodeB) {
		return compose(nodeA, nodeB, null);
	}

	public FSTNode compose(FSTNode nodeA, FSTNode nodeB,
			FSTNode compParent) {

		
		
		System.out.println("COMPOSE\n------------------------------------------------------");
		System.out.print("NodeA: ");
		System.out.println(nodeA.getName());
		System.out.println(nodeA.getType());
		System.out.println(meta.getFeatureName(nodeA));
		System.out.print("NodeB: ");
		System.out.println(nodeB.getName());
		System.out.println(nodeB.getType());
		System.out.println(meta.getFeatureName(nodeB));
		
		System.out.print("compParent: ");
		if (compParent == null) {
			System.out.println("NULL!!");
		} else {
			System.out.println(compParent.getName());
			System.out.println(compParent.getType());
		}
		
		if (nodeA.compatibleWith(nodeB)) {
			FSTNode compNode = nodeA.getShallowClone();
			compNode.setParent(compParent);

			// composed SubTree-stub is integrated in the new Tree, needs
			// children
			if (nodeA instanceof FSTNonTerminal
					&& nodeB instanceof FSTNonTerminal) {
				FSTNonTerminal nonterminalA = (FSTNonTerminal) nodeA;
				FSTNonTerminal nonterminalB = (FSTNonTerminal) nodeB;
				FSTNonTerminal nonterminalComp = (FSTNonTerminal) compNode;

				for (FSTNode childB : nonterminalB.getChildren()) {
					FSTNode childA = nonterminalA.getCompatibleChild(childB);
					// for each child of B get the first compatible child of A
					// (CompatibleChild means a Child which root equals B's
					// root)
					if (childA == null) {
						// no compatible child, FST-node only in B
						nonterminalComp.addChild(rewriteSubtree(childB));
					} else {
						nonterminalComp.addChild(compose(childA, childB,
								nonterminalComp));
					}
				}
				for (FSTNode childA : nonterminalA.getChildren()) {
					FSTNode childB = nonterminalB.getCompatibleChild(childA);
					if (childB == null) {
						// no compatible child, FST-node only in A
						
						nonterminalComp.addChild(rewriteSubtree(childA));
					}
				}
				return nonterminalComp;
			} else if (nodeA instanceof FSTTerminal
					&& nodeB instanceof FSTTerminal
					&& compParent instanceof FSTNonTerminal) {
				FSTTerminal terminalA = (FSTTerminal) nodeA;
				FSTTerminal terminalB = (FSTTerminal) nodeB;
				FSTTerminal terminalComp = (FSTTerminal) compNode;
				FSTNonTerminal nonterminalParent = (FSTNonTerminal) compParent;
				
				CompositionRule applicableRule = null;
				
				//get applicable rule from compositionRules
				for (CompositionRule rule: compositionRules) {
					if (terminalA.getCompositionMechanism().equals(rule.getRuleName())) {						
						 applicableRule = rule;
						 break;
					}	
				}
				
				if (applicableRule != null) {
					//apply composition rule
					applicableRule.compose(terminalA, terminalB, terminalComp, nonterminalParent);
				} else {
					System.err.println("Error: don't know how to compose terminals: "
							+ terminalB.toString()
							+ " replaces "
							+ terminalA.toString());					
				}
				return terminalComp;

				
			}
			return null;
		} else
			return null;
	}


}